import os
import re
from typing import cast

from muxtools import PathLike, ensure_path, error, warn
from jetpytools import CustomValueError
from vstools import vs

from .encoders.types import Zone

__all__ = ["settings_builder_x265", "settings_builder_x264", "sb", "sb265", "sb264"]


def is_full_zone(zone: Zone) -> bool:
    if isinstance(zone[2], str):
        if len(zone) < 4:
            raise error(f"Zone '{zone}' is invalid.")
        return True
    else:
        return False


def norm_zones(clip_or_max_frames: vs.VideoNode | int, zones: Zone | list[Zone] | None) -> list[Zone]:
    """
    Normalize zones to be within the clip's range.

    :param clip_or_max_frames:      The clip or a max frame count to normalize to.
    :param zones:                   The zones to normalize.

    :return:                        The normalized zones.
    """

    if not zones:
        return []

    max_frames = clip_or_max_frames if isinstance(clip_or_max_frames, int) else clip_or_max_frames.num_frames

    if not isinstance(zones, list):
        zones = [zones]

    newzones = list[Zone]()

    for zone in zones:
        start, end, *params = zone

        if start is None:
            start = 0
        elif isinstance(start, int) and start < 0:
            start = max_frames - abs(start)

        if end is None:
            end = max_frames - 1
        elif isinstance(end, int) and end < 0:
            end = max_frames - abs(end)

        if start > end:
            raise CustomValueError(f"Zone '{zone}' start frame is after end frame!", norm_zones, f"{start} > {end}")

        newzones.append(cast(Zone, (start, min(end, max_frames - 1), *params)))  # type:ignore[type-var]

    return newzones


def shift_zones(zones: Zone | list[Zone] | None, start_frame: int = 0) -> list[Zone] | None:
    if not zones:
        return None
    if not isinstance(zones, list):
        zones = [zones]

    newzones = list[Zone]()

    for zone in zones:
        assert zone[0] is not None and zone[1] is not None
        start = zone[0] - start_frame
        end = zone[1] - start_frame
        if end < 0:
            continue
        if start < 0:
            start = 0

        new_zone = list(zone)
        new_zone[0] = start
        new_zone[1] = end
        newzones.append(tuple(new_zone))  # type: ignore

    return newzones


def zones_to_args(zones: Zone | list[Zone] | None, x265: bool) -> list[str]:
    args: list[str] = []
    if not zones:
        return args
    if not isinstance(zones, list):
        zones = [zones]
    zones_settings: str = ""
    for i, zone in enumerate(zones):
        if is_full_zone(zone):
            if x265 and str(zone[2]).lower() not in ["q", "b"]:
                raise error(f"Zone '{zone}' is invalid for x265. Please only use b or q.")
            zones_settings += f"{zone[0]},{zone[1]},{zone[2]}={zone[3]}"  # type: ignore
        else:
            zones_settings += f"{zone[0]},{zone[1]},b={zone[2]}"
        if i != len(zones) - 1:
            zones_settings += "/"
    args.extend(["--zones", zones_settings])
    return args


def settings_builder_x265(
    preset: str | int = "slower",
    crf: float = 14.0,
    qcomp: float = 0.75,
    psy_rd: float = 2.0,
    psy_rdoq: float = 2.0,
    aq_strength: float = 0.75,
    aq_mode: int = 3,
    rd: int = 4,
    rect: bool = True,
    amp: bool = False,
    chroma_qpoffsets: int = -2,
    tu_intra_depth: int = 2,
    tu_inter_depth: int = 2,
    rskip: bool | int = 0,
    tskip: bool = False,
    ref: int = 4,
    bframes: int = 16,
    cutree: bool = False,
    rc_lookahead: int = 60,
    subme: int = 5,
    me: int = 3,
    b_intra: bool = True,
    weightb: bool = True,
    deblock: list[int] | str = [-2, -2],
    append: str = "",
    **kwargs,
) -> str:
    # Simple insert values
    settings = f" --preset {preset} --crf {crf} --bframes {bframes} --ref {ref} --rc-lookahead {rc_lookahead} --subme {subme} --me {me}"
    settings += f" --aq-mode {aq_mode} --aq-strength {aq_strength} --qcomp {qcomp} --cbqpoffs {chroma_qpoffsets} --crqpoffs {chroma_qpoffsets}"
    settings += f" --rd {rd} --psy-rd {psy_rd} --psy-rdoq {psy_rdoq} --tu-intra-depth {tu_intra_depth} --tu-inter-depth {tu_inter_depth}"

    # Less simple
    settings += f" --{'rect' if rect else 'no-rect'} --{'amp' if amp else 'no-amp'} --{'tskip' if tskip else 'no-tskip'}"
    settings += f" --{'b-intra' if b_intra else 'no-b-intra'} --{'weightb' if weightb else 'no-weightb'} --{'cutree' if cutree else 'no-cutree'}"
    settings += f" --rskip {int(rskip) if isinstance(rskip, bool) else rskip}"

    if isinstance(deblock, list):
        deblock = f"{str(deblock[0])}:{str(deblock[1])}"
    settings += f" --deblock={deblock}"

    # Don't need to change these lol
    settings += " --no-sao --no-sao-non-deblock --no-strong-intra-smoothing --no-open-gop"

    for k, v in kwargs.items():
        prefix = "--"
        if k.startswith("_"):
            prefix = "-"
            k = k[1:]
        settings += f" {prefix}{k.replace('_', '-')} {v}"

    settings += (" " + append.strip()) if append.strip() else ""
    return settings


def settings_builder_x264(
    preset: str = "placebo",
    crf: float = 13,
    qcomp: float = 0.7,
    psy_rd: float = 1.0,
    psy_trellis: float = 0.0,
    trellis: int | None = None,
    aq_strength: float = 0.8,
    aq_mode: int = 3,
    ref: int = 16,
    bframes: int = 16,
    mbtree: bool = False,
    rc_lookahead: int = 250,
    me: str = "umh",
    subme: int = 11,
    threads: int = 6,
    merange: int = 32,
    deblock: list[int] | str = [-1, -1],
    dct_decimate: bool = False,
    append: str = "",
    **kwargs,
) -> str:
    # Simple insert values
    settings = f" --preset {preset} --crf {crf} --bframes {bframes} --ref {ref} --rc-lookahead {rc_lookahead} --me {me} --merange {merange}"
    settings += f" --aq-mode {aq_mode} --aq-strength {aq_strength} --qcomp {qcomp}"
    settings += f" --psy-rd {psy_rd}:{psy_trellis} --subme {subme} --threads {threads}"
    if trellis is not None:
        settings += f" --trellis {trellis}"

    # Less simple
    settings += f" {'--no-mbtree' if not mbtree else ''} {'--no-dct-decimate' if not dct_decimate else ''}"

    if isinstance(deblock, list):
        deblock = f"{str(deblock[0])}:{str(deblock[1])}"
    settings += f" --deblock={deblock}"

    for k, v in kwargs.items():
        prefix = "--"
        if k.startswith("_"):
            prefix = "-"
            k = k[1:]
        settings += f" {prefix}{k.replace('_', '-')} {v}"

    settings += (" " + append.strip()) if append.strip() else ""
    return settings


sb = settings_builder_x265
sb265 = sb
sb264 = settings_builder_x264


def file_or_default(file: PathLike | list[str], default: str, no_warn: bool = False) -> tuple[str | list[str], bool]:
    if isinstance(file, list):
        return file, False
    if file is not None and os.path.isfile(file):
        file = ensure_path(file, None)
        if file.exists():
            with open(file, "r") as r:
                settings = str(r.read())
                settings = settings.replace("\n", " ")
                settings = re.sub(r"(?:-o|--output) {clip.+?}", "", settings, flags=re.I).strip()
                return settings, True

    if not no_warn:
        warn("Settings file wasn't found. Using default.", None, 3)
    return default, False



import re

from muxtools import warn
from muxtools.helpers.bsf import BSF_Matrix, BSF_Primaries, BSF_Transfer
from jetpytools import KwargsT
from vstools import ChromaLocation, ColorRange, Matrix, Primaries, Transfer, vs, get_video_format

__all__ = ["props_dict", "fill_props", "props_args", "X264_RANGES", "SVT_AV1_RANGES"]

SVT_AV1_RANGES = ("full", "studio")
X264_RANGES = ("pc", "tv")


def props_dict(
    clip: vs.VideoNode, use_strings: bool, color_range_strings: tuple[str, str] = ("full", "limited"), chromaloc_string: bool = True
) -> dict[str, str]:
    color_range = ColorRange.from_video(clip)
    clip_format = get_video_format(clip)
    props = clip.get_frame(0).props
    chromaloc = ChromaLocation.from_video(clip)
    transfer = BSF_Transfer(Transfer.from_video(clip).value)
    matrix = BSF_Matrix(Matrix.from_video(clip).value)
    primaries = BSF_Primaries(Primaries.from_video(clip).value)

    str_props = KwargsT(
        depth=str(clip_format.bits_per_sample),
        range=color_range_strings[0] if color_range.is_full else color_range_strings[1],
        chromaloc=str(chromaloc.string if use_strings and chromaloc_string else chromaloc.value),
        transfer=str(transfer.value if not use_strings else transfer.name.lower()),
        colormatrix=str(matrix.value if not use_strings else matrix.name.lower()),
        primaries=str(primaries.value if not use_strings else primaries.name.lower()),
        sarnum=str(props.get("_SARNum", 1)),
        sarden=str(props.get("_SARDen", 1)),
        keyint=str(round(clip.fps) * 10),
        min_keyint=str(round(clip.fps)),
        frames=str(clip.num_frames),
        fps_num=str(clip.fps_num),
        fps_den=str(clip.fps_den),
        min_luma=str(16 << (clip_format.bits_per_sample - 8) if color_range.is_limited else 0),
        max_luma=str(235 << (clip_format.bits_per_sample - 8) if color_range.is_limited else (1 << clip_format.bits_per_sample) - 1),
        lookahead=str(min(clip.fps_num * 5, 250)),
    )
    return str_props


def fill_props(settings: str, clip: vs.VideoNode, x265: bool, sar: str | None = None) -> str:
    props = props_dict(clip, False) if x265 else props_dict(clip, True, X264_RANGES, False)
    if sar is not None:
        if not isinstance(sar, str):
            sar = str(sar)
        sarnum = sar if ":" not in sar else sar.split(":")[0]
        sarden = sar if ":" not in sar else sar.split(":")[1]
    else:
        sarnum = props["sarnum"]
        sarden = props["sarden"]
        if sarnum != "1" or sarden != "1":
            warn(f"Are you sure your SAR ({sarnum}:{sarden}) is correct?\nAre you perhaps working on an anamorphic source?", None, 2)
    settings = re.sub(r"{chromaloc(?::.)?}", props["chromaloc"], settings)
    settings = re.sub(r"{primaries(?::.)?}", props["primaries"], settings)
    settings = re.sub(r"{bits(?::.)?}", props["depth"], settings)
    settings = re.sub(r"{matrix(?::.)?}", props["colormatrix"], settings)
    settings = re.sub(r"{range(?::.)?}", props["range"], settings)
    settings = re.sub(r"{transfer(?::.)?}", props["transfer"], settings)
    settings = re.sub(r"{frames(?::.)?}", props["frames"], settings)
    settings = re.sub(r"{fps_num(?::.)?}", props["fps_num"], settings)
    settings = re.sub(r"{fps_den(?::.)?}", props["fps_den"], settings)
    settings = re.sub(r"{min_keyint(?::.)?}", props["min_keyint"], settings)
    settings = re.sub(r"{keyint(?::.)?}", props["keyint"], settings)
    settings = re.sub(r"{sarnum(?::.)?}", sarnum, settings)
    settings = re.sub(r"{sarden(?::.)?}", sarden, settings)
    settings = re.sub(r"{min_luma(?::.)?}", props["min_luma"], settings)
    settings = re.sub(r"{max_luma(?::.)?}", props["max_luma"], settings)
    settings = re.sub(r"{lookahead(?::.)?}", props["lookahead"], settings)
    return settings


def props_args(clip: vs.VideoNode, x265: bool, sar: str | None = None) -> list[str]:
    args: list[str] = []
    props = props_dict(clip, False) if x265 else props_dict(clip, True, X264_RANGES, False)
    if sar is not None:
        if not isinstance(sar, str):
            sar = str(sar)
        sarnum = sar if ":" not in sar else sar.split(":")[0]
        sarden = sar if ":" not in sar else sar.split(":")[1]
    else:
        sarnum = props.get("sarnum")
        sarden = props.get("sarden")
        if sarnum != "1" or sarden != "1":
            warn(f"Are you sure your SAR ({sarnum}:{sarden}) is correct?\nAre you perhaps working on an anamorphic source?", None, 2)

    # fmt: off
    args.extend([
        "--input-depth", props["depth"],
        "--output-depth", props["depth"],
        "--transfer", props["transfer"],
        "--chromaloc", props["chromaloc"],
        "--colormatrix", props["colormatrix"],
        "--range", props["range"],
        "--colorprim", props["primaries"],
        "--sar", f"{sarnum}:{sarden}",
        "--frames", props["frames"]
    ])
    if x265:
        args.extend([
            "--min-luma", props["min_luma"],
            "--max-luma", props["max_luma"]
        ])
    return args
    # fmt: on



import shlex
import subprocess
from vstools import vs, get_video_format
from muxtools import VideoFile, PathLike, make_output, warn, error
from muxtools.utils.dataclass import allow_extra, dataclass

from .base import FFMpegEncoder
from .types import LosslessPreset, ProResProfile

__all__ = ["FFV1", "ProRes"]


@dataclass(config=allow_extra)
class FFV1(FFMpegEncoder):
    """
    Uses ffmpeg to encode a clip to a lossless ffv1 stream.

    :param settings:        Can either be a string of your own settings or any of the 3 presets.
    """

    settings: str | LosslessPreset = LosslessPreset.MIDDLEGROUND

    def encode(self, clip: vs.VideoNode, outfile: PathLike | None = None) -> VideoFile:
        bits = get_video_format(clip).bits_per_sample
        if bits > 10:
            warn(f"You are encoding FFV1 with {bits} bits. It will be massive.", self, 1)
        _base = "-coder 1 -context 0 -g 1 -level 3 -threads 0"
        match self.settings:
            case LosslessPreset.SPEED:
                self.settings = _base + " -slices 30 -slicecrc 0"
            case LosslessPreset.COMPRESSION:
                self.settings = _base + " -slices 16 -slicecrc 1"
            case LosslessPreset.MIDDLEGROUND:
                self.settings = _base + " -slices 24 -slicecrc 1"
            case _:
                self.settings = self.settings

        out = make_output("encoded_ffv1", "mkv", user_passed=outfile)

        input_args, prop_args = self.input_args(clip)
        args = self._default_args() + input_args + ["-c:v", "ffv1"] + prop_args + self.get_custom_args()
        if self.settings:
            args.extend(shlex.split(self.settings))
        args.append(str(out))

        process = subprocess.Popen(args, stdin=subprocess.PIPE)
        self.update_process_affinity(process.pid)
        clip.output(process.stdin, y4m=True)
        process.communicate()

        enc_settings = self.get_mediainfo_settings(shlex.split(self.settings) + self.get_custom_args(), False)
        return VideoFile(out, tags=dict(ENCODER="ffmpeg FFV1", ENCODER_SETTINGS=enc_settings))


@dataclass(config=allow_extra)
class ProRes(FFMpegEncoder):
    """
    Uses ffmpeg's prores-ks encoder to encode a clip to a ProRes stream.\n
    Documentation for params and additional options can be viewed at: https://ffmpeg.org/ffmpeg-codecs.html#Private-Options-for-prores_002dks


    :param profile:         The encoder profile. Basically Quality settings.
                            Chooses the Standard/Default profile for 422 and the '4444' profile for 444 clips if None.
    """

    profile: ProResProfile | int | None = None

    def __post_init__(self):
        if isinstance(self.profile, int) and self.profile not in ProResProfile.__members__.values():
            raise error(f"{self.profile} is not a valid ProRes profile!", self)
        super().__post_init__()

    def encode(self, clip: vs.VideoNode, outfile: PathLike | None = None) -> VideoFile:
        clipf = get_video_format(clip)
        if clipf.subsampling_h != 0:
            raise error("ProRes only supports 422 and 444 subsampling clips.", self)
        if clipf.bits_per_sample > 12:
            raise error("ProRes only supports bitdepths at or below 12.", self)

        profile = self.profile
        if profile is None:
            profile = ProResProfile.DEFAULT if clipf.subsampling_w == 1 else ProResProfile.P4444

        if profile in range(0, 4) and clipf.subsampling_w != 1:
            raise error(f"Profile '{ProResProfile(profile).name}' only supports 422.", self)

        if profile in range(4, 5) and clipf.subsampling_w != 0:
            raise error(f"Profile '{ProResProfile(profile).name}' only supports 444.", self)

        out = make_output("prores", "mkv", user_passed=outfile)
        input_args, prop_args = self.input_args(clip)
        args = self._default_args() + input_args + ["-c:v", "prores_ks"] + prop_args + self.get_custom_args() + ["-profile", str(profile)]
        args.append(str(out))

        process = subprocess.Popen(args, stdin=subprocess.PIPE)
        self.update_process_affinity(process.pid)
        clip.output(process.stdin, y4m=True)
        process.communicate()
        enc_settings = self.get_mediainfo_settings(["-profile", str(profile)] + self.get_custom_args(), False)
        return VideoFile(out, tags=dict(ENCODER="ffmpeg prores_ks", ENCODER_SETTINGS=enc_settings))



import shlex
import subprocess
from vstools import finalize_clip, vs, ChromaLocation
from pathlib import Path
from muxtools import get_executable, VideoFile, PathLike, make_output, warn, get_setup_attr, ensure_path, info, get_workdir, error
from muxtools.utils.env import get_binary_version
from muxtools.utils.dataclass import dataclass, allow_extra

from .base import SupportsQP, VideoEncoder
from .types import LosslessPreset
from ..settings import shift_zones, zones_to_args, norm_zones

from vsmuxtools.utils.source import generate_keyframes, src_file

__all__ = ["x264", "x265", "LosslessX264", "SVTAV1"]


@dataclass(config=allow_extra)
class x264(SupportsQP):
    """
    Encodes your clip to an avc/h264 file using x264.

    :param settings:            This will by default try to look for an `x264_settings` file in your cwd.
                                If it doesn't find one it will warn you and resort to the default settings_builder preset.
                                You can either pass settings as usual or a filepath here.
                                If the filepath doesn't exist it will assume you passed actual settings and pass those to the encoder.

    :param zones:               With this you can tweak settings of specific regions of the video.
                                In x264 this includes but is not limited to CRF.
                                For example (100, 300, "crf", 12) or [(100, 300, "crf", 12), (500, 750, 1.3)]
                                If the third part is not a string it will assume a bitrate multiplier (or "b")

    :param qp_file:             Here you can pass a bool to en/disable or an existing filepath for one.
    :param qp_clip:             Can either be a straight up VideoNode or a SRC_FILE/FileInfo from this package.
                                If neither a clip or a file are given it will simply skip.
                                If only a clip is given it will generate one.

    :param add_props:           This will explicitly add all props taken from the clip to the command line.
                                This will be disabled by default if you are using a file and otherwise enabled.
                                Files can have their own tokens like in vs-encode/vardautomation that will be filled in.

    :param sar:                 Here you can pass your Pixel / Sample Aspect Ratio. This will overwrite whatever is in the clip if passed.
    :param resumable:           Enable or disable resumable encodes. Very useful for people that have scripts that crash their PC (skill issue tbh)
    """

    resumable: bool = True
    x265 = False

    def __post_init__(self):
        self.executable = get_executable("x264")
        self._init_settings(self.x265)

    def _encode_clip(self, clip: vs.VideoNode, out: Path, qpfile: str | None, start_frame: int = 0) -> Path:
        args = [self.executable, "-o", str(out.resolve())]
        if qpfile:
            args.extend(["--qpfile", qpfile])
        if self.settings:
            args.extend(self.settings if isinstance(self.settings, list) else shlex.split(str(self.settings)))
        if self.zones:
            self.zones = norm_zones(clip, self.zones)
            if start_frame:
                self.zones = shift_zones(self.zones, start_frame)
            args.extend(zones_to_args(self.zones, False))
        args.extend(self.get_custom_args() + ["--demuxer", "y4m", "-"])

        process = subprocess.Popen(args, stdin=subprocess.PIPE)
        self.update_process_affinity(process.pid)
        clip.output(process.stdin, y4m=True)
        process.communicate()
        return out


@dataclass(config=allow_extra)
class x265(SupportsQP):
    """
    Encodes your clip to an hevc/h265 file using x265.

    :param settings:            This will by default try to look for an `x265_settings` file in your cwd.
                                If it doesn't find one it will warn you and resort to the default settings_builder preset.
                                You can either pass settings as usual or a filepath here.
                                If the filepath doesn't exist it will assume you passed actual settings and pass those to the encoder.

    :param zones:               With this you can tweak settings of specific regions of the video.
                                In x265 you're basically limited to a flat bitrate multiplier or force QP ("q")
                                For example (100, 300, "b", 1.2) or [(100, 300, "q", 12), (500, 750, 1.3)]
                                If the third part is not a string it will assume a bitrate multiplier (or "b")

    :param qp_file:             Here you can pass a bool to en/disable or an existing filepath for one.
    :param qp_clip:             Can either be a straight up VideoNode or a SRC_FILE/FileInfo from this package.
                                If neither a clip or a file are given it will simply skip.
                                If only a clip is given it will generate one.

    :param add_props:           This will explicitly add all props taken from the clip to the command line.
                                This will be disabled by default if you are using a file and otherwise enabled.
                                Files can have their own tokens like in vs-encode/vardautomation that will be filled in.

    :param sar:                 Here you can pass your Pixel / Sample Aspect Ratio. This will overwrite whatever is in the clip if passed.
    :param resumable:           Enable or disable resumable encodes. Very useful for people that have scripts that crash their PC (skill issue tbh)
    :param csv:                 Either a bool to enable or disable csv logging or a Filepath for said csv.
    """

    resumable: bool = True
    csv: bool | PathLike = True
    x265 = True

    def __post_init__(self):
        self.executable = get_executable("x265")
        self._init_settings(self.x265)

    def _encode_clip(self, clip: vs.VideoNode, out: Path, qpfile: str | None, start_frame: int = 0) -> Path:
        args = [self.executable, "-o", str(out.resolve())]
        if self.csv:
            if isinstance(self.csv, bool):
                show_name = get_setup_attr("show_name", "")
                csv_file = Path(show_name + f"{'_' if show_name else ''}log_x265.csv").resolve()
            else:
                csv_file = ensure_path(self.csv, self)
            args.extend(["--csv", str(csv_file)])
        if qpfile:
            args.extend(["--qpfile", qpfile])
        if self.settings:
            args.extend(self.settings if isinstance(self.settings, list) else shlex.split(str(self.settings)))
        if self.zones:
            self.zones = norm_zones(clip, self.zones)
            if start_frame:
                self.zones = shift_zones(self.zones, start_frame)
            args.extend(zones_to_args(self.zones, True))
        args.extend(self.get_custom_args() + ["--y4m", "--input", "-"])

        process = subprocess.Popen(args, stdin=subprocess.PIPE)
        self.update_process_affinity(process.pid)
        clip.output(process.stdin, y4m=True)
        process.communicate()
        return out


@dataclass(config=allow_extra)
class LosslessX264(VideoEncoder):
    """
    Uses x264 to encode clip to a lossless avc stream.

    :param preset:          Can either be a string of some x264 preset or any of the 3 predefined presets.
    :param settings:        Any other settings you might want to pass. Entirely optional.
    :param add_props:       This will explicitly add all props taken from the clip to the command line.
    """

    preset: str | LosslessPreset = LosslessPreset.MIDDLEGROUND
    settings: str | None = None
    add_props: bool = True

    def encode(self, clip: vs.VideoNode, outfile: PathLike | None = None) -> VideoFile:
        out = make_output("lossless", "264", user_passed=outfile)
        match self.preset:
            case LosslessPreset.SPEED:
                preset = "ultrafast"
            case LosslessPreset.COMPRESSION:
                preset = "veryslow"
            case LosslessPreset.MIDDLEGROUND:
                preset = "medium"
            case _:
                preset = self.preset
        settings = ["--preset", preset, "--qp", "0"] + self.get_custom_args()
        if clip.format.bits_per_sample > 10:
            warn("This encoder does not support a bit depth over 10.\nClip will be dithered to 10 bit.", self, 2)
            clip = finalize_clip(clip, 10)

        if self.settings:
            settings.extend(shlex.split(self.settings))
        avc = x264(shlex.join(settings), add_props=self.add_props, resumable=False)
        avc._update_settings(clip, False)
        avc._encode_clip(clip, out, None, 0)
        return VideoFile(out)


@dataclass(config=allow_extra)
class SVTAV1(VideoEncoder):
    """
    Uses SVtAv1EncApp to encode clip to a av1 stream.\n
    Do not use this for high fidelity encoding.\n
    For better explanations of params, check the `--help` of the encoder or the gitlab wiki page.

    :param preset:          Encoder preset. Lower = slower & better
                            The range is -1 to 13 for the regular SVTAV1 and -3 to 13 for SVTAV1-PSY
    :param crf:             Constant rate factor, lower = better
    :param tune:            The tuning metric. None = 2 for SVTAV1 and 3 for SVTAV1-PSY

    :param qp_clip:         Can either be a straight up VideoNode or a SRC_FILE/FileInfo from this package.
                            It is highly recommended to do this so you can force keyframes.
    """

    preset: int = 4
    crf: int | float = 15
    tune: int | None = None
    qp_clip: vs.VideoNode | src_file | None = None

    def __post_init__(self):
        self.executable = get_executable("SvtAv1EncApp")
        if self.get_process_affinity() is False:
            self.affinity = []
        if not self.qp_clip:
            warn("It is highly recommended to force keyframes with this encoder!\nPlease pass a qp_clip param.", self, 2)

    def _make_keyframes_config(self, clip: vs.VideoNode) -> Path | bool:
        out = get_workdir() / "svt_keyframes.cfg"
        if out.exists():
            info("Reusing existing keyframes config.", self)
            return out
        info("Generating keyframes config file...", self)

        keyframes = generate_keyframes(clip)
        if not keyframes:
            return False
        keyframes_str = f"ForceKeyFrames : {'f,'.join([str(i) for i in keyframes])}f"
        with open(out, "w", encoding="utf-8") as f:
            f.write(keyframes_str)

        info("Done", self)
        return out

    def encode(self, clip: vs.VideoNode, outfile: PathLike | None = None) -> VideoFile:
        from vsmuxtools.video.clip_metadata import props_dict, SVT_AV1_RANGES

        clip_props = props_dict(clip, False, SVT_AV1_RANGES)
        match int(clip_props["chromaloc"]):
            case ChromaLocation.LEFT.value:
                cloc = "left"
            case ChromaLocation.TOP_LEFT.value:
                cloc = "topleft"
            case _:
                raise error("AV1 only supports LEFT and TOPLEFT chroma locations!", self)

        output = make_output("svtav1", ext="ivf", user_passed=outfile)
        encoder = get_binary_version(self.executable, r"(SVT-AV1.+?(?:v\d+.\d+.\d[^ ]+|[0-9a-f]{8,40}))", ["--version"])
        assert encoder

        tags = dict[str, str](ENCODER=encoder)
        args = [self.executable, "-i", "-", "--output", str(output), "--preset", str(self.preset)]
        if self.qp_clip:
            qp_clip = self.qp_clip if isinstance(self.qp_clip, vs.VideoNode) else self.qp_clip.src_cut
            keyframes_config = self._make_keyframes_config(qp_clip)
            if keyframes_config:
                args.extend(["--keyint", "-1", "-c", str(keyframes_config)])
            else:
                warn("No keyframes found.", self)

        if self.crf:
            args.extend(["--crf", str(self.crf)])

        if self.tune is None:
            self.tune = 3 if "psy" in encoder.lower() else 2

        args.extend(["--tune", str(self.tune)])

        # fmt:off
        args.extend(self.get_custom_args() + [
            "--fps-num", clip_props.get("fps_num"),
            "--fps-denom", clip_props.get("fps_den"),
            "--input-depth", clip_props.get("depth"),
            "--chroma-sample-position", cloc,
            "--color-primaries", clip_props.get("primaries"),
            "--transfer-characteristics", clip_props.get("transfer"),
            "--matrix-coefficients", clip_props.get("colormatrix"),
            "--color-range", clip_props.get("range")
        ])
        # fmt: on

        process = subprocess.Popen(args, stdin=subprocess.PIPE)
        self.update_process_affinity(process.pid)
        clip.output(process.stdin, y4m=True)
        process.communicate()
        tags.update(ENCODER_SETTINGS=self.get_mediainfo_settings(args))
        return VideoFile(output, tags=tags)



from fractions import Fraction
from muxtools import parse_chapters_bdmv, PathLike, GlobSearch, Chapter, Chapters as Ch, error, resolve_timesource_and_scale
from muxtools.utils.types import TimeScale, TimeScaleT, TimeSourceT

from ..utils.source import src_file

__all__ = ["Chapters"]


class Chapters(Ch):
    def __init__(
        self,
        chapter_source: src_file | PathLike | GlobSearch | Chapter | list[Chapter],
        timesource: TimeSourceT = None,
        timescale: TimeScaleT = TimeScale.MKV,
        _print: bool = True,
    ) -> None:
        """
        Convenience class for chapters

        :param chapter_source:      Input either src_file/FileInfo, txt with ogm chapters, xml or (a list of) self defined chapters.
        :param timesource:          The source of timestamps/timecodes. For details check the docstring on the type.\n
                                    Will be taken from input if it's a src_file and assume the usual 24 if not.

        :param timescale:           Unit of time (in seconds) in terms of which frame timestamps are represented.\n
                                    For details check the docstring on the type.

        :param _print:              Prints chapters after parsing and after trimming.
        """
        if isinstance(chapter_source, src_file):
            if isinstance(chapter_source.file, list):
                # I'll make a workaround for this soonish
                raise error("Cannot currently parse chapters when splicing multiple files.", self)
            clip_fps = Fraction(chapter_source.src.fps_num, chapter_source.src.fps_den)
            self.timestamps = resolve_timesource_and_scale(timesource if timesource else clip_fps, timescale, caller=self)
            self.chapters = parse_chapters_bdmv(chapter_source.file, clip_fps, chapter_source.src_cut.num_frames, _print)
            if self.chapters and chapter_source.trim:
                self.trim(chapter_source.trim[0] or 0, chapter_source.trim[1] or 0, chapter_source.src_cut.num_frames)
                if _print:
                    print("After trim:")
                    self.print()
        else:
            # TODO: Why does this fail too?
            super().__init__(chapter_source, timesource or Fraction(24000, 1001), timescale, _print)  # type: ignore



from enum import IntEnum
from typing import Union

__all__ = ["LosslessPreset", "ProResProfile"]

ZoneFrame = Union[int, None]
Zone = tuple[ZoneFrame, ZoneFrame, float | str] | tuple[ZoneFrame, ZoneFrame, str, float | int | str]


class LosslessPreset(IntEnum):
    SPEED = 1
    COMPRESSION = 2
    MIDDLEGROUND = 3


class ProResProfile(IntEnum):
    AUTO = -1
    PROXY = 0
    LT = 1
    DEFAULT = 2
    HQ = 3
    P4444 = 4
    P4444XQ = 5



from . import utils, video, extension

from .utils import *
from .video import *

# TODO: Figure out something for the redefinition of modules here
from muxtools.main import *
from muxtools.audio import *  # type: ignore
from muxtools.muxing import *  # type: ignore
from muxtools.subtitle import *  # type: ignore
from muxtools.utils import *  # type: ignore

from .extension import *  # type: ignore

__version__: str
__version_tuple__: tuple[int | str, ...]

try:
    from ._version import __version__, __version_tuple__
except ImportError:
    __version__ = "0.0.0+unknown"
    __version_tuple__ = (0, 0, 0, "+unknown")



import shlex
from vstools import finalize_clip, vs, get_video_format
from pathlib import Path
from dataclasses import dataclass
from abc import ABC, abstractmethod
from muxtools import ensure_path_exists, VideoFile, PathLike, make_output, info, warn, error, get_executable
from muxtools.utils.dataclass import CLIKwargs

from .types import Zone
from ..resumable import merge_parts, parse_keyframes
from ..settings import file_or_default, sb264, sb265
from ..clip_metadata import fill_props, props_args, props_dict
from vsmuxtools.utils.source import generate_qp_file, src_file

__all__ = ["VideoEncoder", "FFMpegEncoder", "SupportsQP"]


@dataclass
class VideoEncoder(CLIKwargs, ABC):
    resumable = False

    @abstractmethod
    def encode(self, clip: vs.VideoNode, outfile: PathLike | None = None) -> VideoFile:
        """
        To actually run the encode.

        :param clip:            Your videonode
        :param outfile:         Can be a custom output file or directory.
                                The correct extension will automatically be appended.

        Returns a VideoFile object.
        If you're only interested in the path you can just do `VideoFile.file`.
        """
        ...

    def _update_progress(self, current_frame, total_frames):
        print(f"\rVapoursynth: {current_frame} / {total_frames} ({100 * current_frame // total_frames}%) || Encoder: ", end="")


@dataclass
class FFMpegEncoder(VideoEncoder, ABC):
    def __post_init__(self):
        self.executable = get_executable("ffmpeg")

    def _default_args(self) -> list[str]:
        return [self.executable, "-v", "quiet", "-stats"]

    def _pixfmt_for_clip(self, clip: vs.VideoNode) -> str:
        videoformat = get_video_format(clip)
        if videoformat.color_family != vs.YUV:
            raise error("Only YUV input allowed for FFMPEG pipes/encoders!", self)

        allowed_depths = range(8, 18, 2)
        if videoformat.bits_per_sample not in allowed_depths:
            raise error(f"Only the following bitdepths are allowed: {', '.join([str(x) for x in allowed_depths])}", self)

        formatname = videoformat.name.replace("P8", "P").lower()
        return formatname + "le" if videoformat.bits_per_sample > 8 else formatname

    def input_args(self, clip: vs.VideoNode) -> tuple[list[str], list[str]]:
        props = props_dict(clip, True)
        # fmt: off
        prop_args = [
            "-r", f"{props['fps_num']}/{props['fps_den']}",
            "-color_range", props["range"],
            "-colorspace", props["colormatrix"],
            "-color_primaries", props["primaries"],
            "-color_trc", props["transfer"],
            "-chroma_sample_location", props["chromaloc"],
        ]
        input_arguments = prop_args + ["-f", "yuv4mpegpipe", "-i", "-"] + ["-pix_fmt", self._pixfmt_for_clip(clip)]
        return input_arguments, prop_args
        # fmt: on


@dataclass
class SupportsQP(VideoEncoder):
    settings: str | PathLike | list[str] | None = None
    zones: Zone | list[Zone] | None = None
    qp_file: PathLike | bool | None = None
    qp_clip: src_file | vs.VideoNode | None = None
    add_props: bool | None = None
    sar: str | None = None
    quiet_merging: bool = True
    x265 = True

    def _get_qpfile(self, start_frame: int = 0) -> str:
        if not self.qp_file and not self.qp_clip:
            return ""

        if not isinstance(self.qp_file, bool) and self.qp_file is not None:
            return str(ensure_path_exists(self.qp_file, self).resolve())

        if self.qp_clip:
            if isinstance(self.qp_clip, src_file):
                self.qp_clip = self.qp_clip.src_cut
            return generate_qp_file(self.qp_clip, start_frame)

        return ""

    def _init_settings(self, x265: bool):
        if not self.settings:
            s, p = file_or_default(f"{'x265' if x265 else 'x264'}_settings", sb265() if x265 else sb264())
            self.was_file = p
            self.settings = s
        else:
            s, p = file_or_default(self.settings, str(self.settings), True)
            self.was_file = p
            self.settings = s

        if self.add_props is None:
            self.add_props = not getattr(self, "was_file", False)

    def _update_settings(self, clip: vs.VideoNode, x265: bool):
        if self.was_file:
            self.settings = fill_props(str(self.settings), clip, x265, self.sar)

        self.settings = self.settings if isinstance(self.settings, list) else shlex.split(str(self.settings))

        if self.add_props:
            self.settings.extend(props_args(clip, x265, self.sar))

    @abstractmethod
    def _encode_clip(self, clip: vs.VideoNode, out: Path, qpfile: str | None, start_frame: int = 0) -> Path: ...

    def encode(self, clip: vs.VideoNode, outfile: PathLike | None = None) -> VideoFile:
        if clip.format.bits_per_sample > (12 if self.x265 else 10):
            warn(f"This encoder does not support a bit depth over {(12 if self.x265 else 10)}.\nClip will be dithered to 10 bit.", self, 2)
            clip = finalize_clip(clip, 10)
        self._update_settings(clip, self.x265)
        out = make_output(
            ensure_path_exists(self.qp_clip.file, self).stem if isinstance(self.qp_clip, src_file) else "encoded",
            "265" if self.x265 else "264",
            "encoded" if isinstance(self.qp_clip, src_file) else "",
            outfile,
        )
        if not self.resumable:
            return VideoFile(self._encode_clip(clip, out, self._get_qpfile()))

        pattern = out.with_stem(out.stem + "_part_???")
        parts = sorted(pattern.parent.glob(pattern.name))
        info(f"Found {len(parts)} part{'s' if len(parts) != 1 else ''} for this encode")

        keyframes = list[int]()
        for i, p in enumerate(parts):
            try:
                info(f"Parsing keyframes for part {i}...")
                kf = parse_keyframes(p)[-1]
                if kf == 0:
                    del parts[-1]
                else:
                    keyframes.append(kf)
            except:
                del parts[-1]
        fout = out.with_stem(out.stem + f"_part_{len(parts):03.0f}")
        start_frame = sum(keyframes)
        info(f"Starting encode at frame {start_frame}")

        # TODO: Normalize and adjust existing zones to the new start frame

        clip = clip[start_frame:]
        self._encode_clip(clip, fout, self._get_qpfile(start_frame), start_frame)

        info("Remuxing and merging parts...")
        merge_parts(fout, out, keyframes, parts, self.quiet_merging)
        return VideoFile(out, source=ensure_path_exists(self.qp_clip.file, self) if isinstance(self.qp_clip, src_file) else None)



from vstools import vs
from fractions import Fraction
from datetime import timedelta
from typing_extensions import Self

from muxtools import (
    PathLike,
    ensure_path_exists,
    debug,
    ParsedFile,
    TrackType,
    VideoFile,
    VideoTrack,
    MkvTrack,
    TimeScale,
    resolve_timesource_and_scale,
    get_timemeta_from_video,
    TimeType,
    error,
    warn,
)
from muxtools import SubFile as MTSubFile, SubFilePGS as MTSubFilePGS
from muxtools.subtitle import _Line
from muxtools.subtitle.sub import LINES
from muxtools.utils.types import TimeSourceT, TimeScaleT

from ..utils.source import src_file

__all__ = ["SubFile", "SubFilePGS"]


class SubFile(MTSubFile):
    """
    Utility class representing an ASS/SSA subtitle file with various functions to run on.

    :param file:            Can be a string, Path object or GlobSearch.
                            If the GlobSearch returns multiple results or if a list was passed it will merge them.

    :param container_delay: Set a container delay used in the muxing process later.
    :param source:          The file this sub originates from, will be set by the constructor.
    :param encoding:        Encoding used for reading and writing the subtitle files.
    """

    def truncate_by_video(
        self,
        source: PathLike | VideoTrack | MkvTrack | VideoFile | vs.VideoNode,
        timesource: TimeSourceT = None,
        timescale: TimeScaleT = TimeScale.MKV,
    ) -> Self:
        """
        Removes lines that start after the video ends and trims lines that extend past it.

        :param source:          Can be any video file or a VideoNode
        :param timesource:      The source of timestamps/timecodes. For details check the docstring on the type.
        :param timescale:       Unit of time (in seconds) in terms of which frame timestamps are represented.\n
                                For details check the docstring on the type.
        """

        if isinstance(source, vs.VideoNode):
            frames = source.num_frames
            if not timesource:
                timesource = Fraction(source.fps_num, source.fps_den)
            resolved_ts = resolve_timesource_and_scale(timesource, timescale, allow_warn=False, caller=self)
        else:
            if isinstance(source, VideoTrack) or isinstance(source, MkvTrack) or isinstance(source, VideoFile):
                file = ensure_path_exists(source.file, self)
            else:
                file = ensure_path_exists(source, self)

            parsed = ParsedFile.from_file(file, self)

            assert parsed.find_tracks(type=TrackType.VIDEO, error_if_empty=True, caller=self)[0]

            meta = get_timemeta_from_video(file, caller=self)
            frames = len(meta.pts)
            if not timesource:
                timesource = meta
                timescale = meta.timescale
            resolved_ts = resolve_timesource_and_scale(timesource, timescale, allow_warn=False, caller=self)

        cutoff = timedelta(milliseconds=resolved_ts.frame_to_time(frames + 1, TimeType.START, 2, True) * 10)

        def filter_lines(lines: LINES):
            removed = 0
            trimmed = 0
            new_list = list[_Line]()
            for line in lines:
                if line.start > cutoff:
                    removed += 1
                    continue
                if line.end > cutoff:
                    line.end = timedelta(milliseconds=resolved_ts.frame_to_time(frames, TimeType.END, 2, True) * 10)
                    trimmed += 1
                new_list.append(line)

            if removed or trimmed:
                if removed:
                    debug(f"Removed {removed} line{'s' if removed != 1 else ''} that started past the video", self)
                if trimmed:
                    debug(f"Trimmed {trimmed} line{'s' if trimmed != 1 else ''} that extended past the video", self)

            return new_list

        return self.manipulate_lines(filter_lines)


class SubFilePGS(MTSubFilePGS):
    """
    Utility class representing a PGS/SUP subtitle file.

    :param file:            Can be a string, Path object or GlobSearch.
    :param container_delay: Set a container delay used in the muxing process later.
    :param source:          The file this sub originates from, will be set by the constructor.
    """

    @classmethod
    def extract_from(cls: type[Self], fileIn: PathLike | src_file, track: int = 0, preserve_delay: bool = False, quiet: bool = True) -> Self:
        """
        Extract a PGS subtitle track from a file using ffmpeg.\n

        :param fileIn:          The input file to extract from.
        :param track:           The track number to extract.
        :param preserve_delay:  If True, the container delay will be preserved.
        :param quiet:           If True, suppresses ffmpeg output.
        :return:                An instance of SubFilePGS containing the extracted subtitle.
        """
        f = fileIn
        if isinstance(f, src_file):
            if isinstance(f.file, list):
                # I'll make a workaround for this soonish
                raise error("Cannot currently parse chapters when splicing multiple files.", cls.__name__)
            f = f.file
        new = super().extract_from(
            f,
            track=track,
            preserve_delay=preserve_delay,
            quiet=quiet,
        )
        if isinstance(fileIn, src_file) and fileIn.trim:
            if fileIn.trim[0]:
                debug(f"Shifting extracted subtitle by -{fileIn.trim[0]} frames...", cls.__name__)
                new = new.shift(-fileIn.trim[0], timesource=fileIn.file, quiet=quiet)
            if fileIn.trim[1]:
                warn("Trimming is currently not supported for PGS subtitles.", cls.__name__)
        return new



from typing import Sequence
from collections.abc import Callable
from vstools import vs, get_video_format, depth

from .base import VideoEncoder
from .ffmpeg import ProRes
from .types import ProResProfile

from muxtools import VideoFile, PathLike, error, get_workdir
from muxtools.utils.dataclass import dataclass, allow_extra

__all__ = ["IntermediaryEncoder", "ProResIntermediary"]


VS_FUNCTION = Callable[[vs.VideoNode], vs.VideoNode]


@dataclass(config=allow_extra)
class IntermediaryEncoder:
    """
    Encoder that will create an intermediary first and then encode that intermediary to the target encoders.

    :param encoder:             The intermediary encoder. Might recommend prores or ffv1.

    :param target_encoders:     Target encoders to use. You can also pass a tuple with a function to call on the clips before running.
                                For example: (x265(), lambda clip: clip.nlm_cuda.NLMeans(h=2.0))

    :param indexer:             Here you can pass a custom indexing function ala FileInfo. Uses vsmuxtools.src otherwise.
    """

    encoder: VideoEncoder
    target_encoders: Sequence[VideoEncoder | tuple[VideoEncoder, VS_FUNCTION]]
    indexer: Callable[[str], vs.VideoNode] | None = None

    def encode(self, clip: vs.VideoNode, outfile: PathLike | None = None) -> list[VideoFile]:
        intermediary = self.encoder.encode(clip, get_workdir() / "intermediary")
        from vsmuxtools import src as src_index

        index_clip = self.indexer(str(intermediary.file)) if self.indexer else src_index(intermediary.file)

        outputs = list[VideoFile]()

        for target in self.target_encoders:
            temp_clip = temp_clip = target[1](index_clip) if isinstance(target, tuple) else index_clip
            encoder = target[0] if isinstance(target, tuple) else target
            result = encoder.encode(temp_clip)
            outputs.append(result)

        return outputs


@dataclass(config=allow_extra)
class ProResIntermediary:
    """
    This encodes to prores first and will upscale chroma to 422 with point if needed and undo it before passing to other encoders.

    :param target_encoders:     Target encoders to use. You can also pass a tuple with a function to call on the clips before running.
                                For example: (x265(), lambda clip: clip.nlm_cuda.NLMeans(h=2.0))

    :param indexer:             Here you can pass a custom indexing function ala FileInfo. Uses vsmuxtools.src otherwise.

    :param profile:             The encoder profile. Basically Quality settings.
                                Chooses the Standard/Default profile for 422 and the '4444' profile for 444 clips if None.
    """

    target_encoders: Sequence[VideoEncoder]
    indexer: Callable[[str], vs.VideoNode] | None = None
    profile: ProResProfile | int | None = None

    def encode(self, clip: vs.VideoNode, outfile: PathLike | None = None) -> list[VideoFile]:
        clipf = get_video_format(clip)
        try:
            from vskernels import Point
        except:
            raise error("You need to install vskernels for this", self)
        if clipf.subsampling_h != 0:
            clip = Point().resample(clip, format=clipf.replace(subsampling_h=0))

        targets_with_resample = [(enc, lambda x: depth(Point().resample(x, clipf), 10)) for enc in self.target_encoders]

        encoder = IntermediaryEncoder(
            ProRes(self.profile),
            self.target_encoders if clipf.subsampling_h == 0 else targets_with_resample,
            self.indexer,
        )
        return encoder.encode(clip)



from pathlib import Path
from typing import Callable, Sequence, TYPE_CHECKING, MutableMapping, Any
from fractions import Fraction
from jetpytools import KwargsT
from enum import IntEnum
from vstools import (
    vs,
    core,
    initialize_clip,
    ColorRangeLike,
    MatrixLike,
    TransferLike,
    PrimariesLike,
    DitherType,
    FieldBasedLike,
    ChromaLocationLike,
    vs_object,
)
from muxtools import (
    Trim,
    PathLike,
    parse_m2ts_path,
    ensure_path_exists,
    info,
    get_workdir,
    get_temp_workdir,
    clean_temp_files,
    TrackType,
    GlobSearch,
    error,
    debug,
    warn,
    ParsedFile,
)

from muxtools.audio.preprocess import classproperty


__all__ = ["src_file", "SRC_FILE", "FileInfo", "src", "frames_to_samples", "f2s", "SourceFilter"]


class SourceFilter(IntEnum):
    """Filter used to index the source file"""

    AUTO = 0
    """
    LSMASH for m2ts regardless of previewing; FFMS2 if previewing, BESTSOURCE otherwise.
    """
    BESTSOURCE = 1
    """
    Slow initial indexing but guaranteed accuracy.
    https://github.com/vapoursynth/bestsource
    """
    FFMS2 = 2
    """
    Fast, generally good enough.
    https://github.com/FFMS/ffms2
    """
    LSMASH = 3
    """
    Also fast but notably the only thing that isn't affected by a certain m2ts ffmpeg bug.\n
    May be deprecated once a new FFMS2 build is available.\n
    https://github.com/HomeOfAviSynthPlusEvolution/L-SMASH-Works
    """

    BS = BESTSOURCE
    """Alias for BESTSOURCE"""


class src_file(vs_object):
    file: Path | list[Path]
    trim: Trim | None = None
    preview_sourcefilter: SourceFilter | None
    sourcefilter: SourceFilter
    idx: Callable[[str], vs.VideoNode] | None = None
    idx_args: dict[str, Any] = {}

    def __init__(
        self,
        file: PathLike | GlobSearch | Sequence[PathLike],
        trim: Trim | None = None,
        preview_sourcefilter: SourceFilter | None = SourceFilter.FFMS2,
        sourcefilter: SourceFilter = SourceFilter.BESTSOURCE,
        idx: Callable[[str], vs.VideoNode] | None = None,
        **kwargs,
    ):
        """
        Custom `FileInfo` kind of thing for convenience

        :param file:                    Either a string based filepath or a Path object
        :param trim:                    Can be a single trim or a sequence of trims.
        :param preview_sourcefilter:    Source filter to be used when previewing using vspreview.
                                        `None` will make it fall back to `sourcefilter`.
        :param sourcefilter:            Source filter to be used otherwise.
        :param idx:                     Custom indexer for the input file. Pass a function that takes a string in and returns a vs.VideoNode.
        """
        if isinstance(file, Sequence) and not isinstance(file, str) and len(file) == 1:
            file = file[0]

        self.file = (
            [ensure_path_exists(f, self) for f in file]
            if isinstance(file, Sequence) and not isinstance(file, str)
            else ensure_path_exists(file, self)
        )
        self.trim = trim
        self.preview_sourcefilter = preview_sourcefilter
        self.sourcefilter = sourcefilter
        self.idx = idx
        self.idx_args = kwargs

    def __call_indexer(self, fileIn: Path):
        if self.idx:
            return self.idx(str(fileIn.resolve()))
        else:
            return src(fileIn, preview_sourcefilter=self.preview_sourcefilter, sourcefilter=self.sourcefilter, **self.idx_args)

    def __index_clip(self):
        if isinstance(self.file, list):
            indexed = core.std.Splice([self.__call_indexer(f) for f in self.file])
        else:
            indexed = self.__call_indexer(self.file)
        cut = indexed
        if self.trim:
            self.trim = list(self.trim)
            if self.trim[0] is None:
                self.trim[0] = 0
            if self.trim[1] is None or self.trim[1] == 0:
                if self.trim[0] < 0:
                    cut = (indexed[0] * abs(self.trim[0])) + indexed
                else:
                    cut = indexed[self.trim[0] :]
            else:
                if self.trim[0] < 0:
                    cut = (indexed[0] * abs(self.trim[0])) + indexed[: self.trim[1]]
                else:
                    cut = indexed[self.trim[0] : self.trim[1]]
            self.trim = tuple(self.trim)

        if not isinstance(self.file, list) and self.file.suffix.lower() == ".dgi":
            if self.file.with_suffix(".m2ts").exists():
                self.file = self.file.with_suffix(".m2ts")
            else:
                self.file = parse_m2ts_path(self.file)

        setattr(self, "clip", indexed)
        setattr(self, "clip_cut", cut)

    @property
    def src(self) -> vs.VideoNode:
        if not hasattr(self, "clip"):
            self.__index_clip()
        return self.clip  # type: ignore

    @property
    def src_cut(self) -> vs.VideoNode:
        if not hasattr(self, "clip_cut"):
            self.__index_clip()
        return self.clip_cut  # type: ignore

    def init(
        self,
        bits: int | None = None,
        matrix: MatrixLike | None = None,
        transfer: TransferLike | None = None,
        primaries: PrimariesLike | None = None,
        chroma_location: ChromaLocationLike | None = None,
        color_range: ColorRangeLike | None = None,
        field_based: FieldBasedLike | None = None,
        strict: bool = False,
        dither_type: DitherType = DitherType.AUTO,
    ) -> vs.VideoNode:
        """
        Getter that calls `vstools.initialize_clip` on the src clip for convenience
        """
        return initialize_clip(
            self.src, bits, matrix, transfer, primaries, chroma_location, color_range, field_based, strict, dither_type, func=self.init
        )

    def init_cut(
        self,
        bits: int | None = None,
        matrix: MatrixLike | None = None,
        transfer: TransferLike | None = None,
        primaries: PrimariesLike | None = None,
        chroma_location: ChromaLocationLike | None = None,
        color_range: ColorRangeLike | None = None,
        field_based: FieldBasedLike | None = None,
        strict: bool = False,
        dither_type: DitherType = DitherType.AUTO,
    ) -> vs.VideoNode:
        """
        Getter that calls `vstools.initialize_clip` on the src_cut clip for convenience
        """
        return initialize_clip(
            self.src_cut, bits, matrix, transfer, primaries, chroma_location, color_range, field_based, strict, dither_type, func=self.init_cut
        )

    def get_audio(self, track: int = 0, **kwargs) -> vs.AudioNode:
        """
        Indexes the specified audio track from the input file(s).
        """
        file = self.file if isinstance(self.file, list) else [self.file]

        nodes = list[vs.AudioNode]()
        for f in file:
            parsed = ParsedFile.from_file(f, self)
            absolute = parsed.find_tracks(relative_id=track, type=TrackType.AUDIO, error_if_empty=True, caller=self)[0].index
            nodes.append(core.bs.AudioSource(str(f.resolve()), absolute, **kwargs))

        return nodes[0] if len(nodes) == 1 else core.std.AudioSplice(nodes)

    def get_audio_trimmed(self, track: int = 0, **kwargs) -> vs.AudioNode:
        """
        Gets the indexed audio track with the trim specified in the src_file.
        """
        node = self.get_audio(track, **kwargs)
        if self.trim:
            if self.trim[1] is None or self.trim[1] == 0:
                node = node[f2s(self.trim[0], node, self.src) :]
            else:
                node = node[f2s(self.trim[0], node, self.src) : f2s(self.trim[1], node, self.src)]
        return node

    def __vs_del__(self, core_id: int) -> None:
        if not TYPE_CHECKING:
            setattr(self, "clip", None)
            setattr(self, "clip_cut", None)

            for v in self.__dict__.values():
                if not isinstance(v, MutableMapping):
                    continue

                for k2, v2 in v.items():
                    if isinstance(v2, vs.VideoNode):
                        v[k2] = None

    @staticmethod
    def BDMV(
        root_dir: PathLike,
        playlist: int = 0,
        entries: int | list[int] | Trim | None = None,
        angle: int = 0,
        trim: Trim | None = None,
        preview_sourcefilter: SourceFilter | None = SourceFilter.AUTO,
        sourcefilter: SourceFilter = SourceFilter.BESTSOURCE,
        idx: Callable[[str], vs.VideoNode] | None = None,
        **kwargs: KwargsT,
    ) -> "src_file":
        root_dir = ensure_path_exists(root_dir, "BDMV", True)
        mpls = core.mpls.Read(str(root_dir), playlist, angle)
        playlist_clips: list[str] = mpls["clip"]
        if entries is not None:
            if isinstance(entries, int):
                clips = playlist_clips[entries]
            elif isinstance(entries, list):
                clips = [file for index, file in enumerate(playlist_clips) if index in entries]
            else:
                if entries[0] is None and entries[1]:
                    clips = playlist_clips[: entries[1]]
                elif entries[1] is None:
                    clips = playlist_clips[entries[0] :]
                else:
                    clips = playlist_clips[entries[0] : entries[1]]
        else:
            clips = playlist_clips
        return src_file(clips, trim, preview_sourcefilter, sourcefilter, idx, **kwargs)

    @classproperty
    def AUTO(self) -> SourceFilter:
        return SourceFilter.AUTO

    @classproperty
    def BESTSOURCE(self) -> SourceFilter:
        return SourceFilter.BESTSOURCE

    @classproperty
    def BS(self) -> SourceFilter:
        return SourceFilter.BESTSOURCE

    @classproperty
    def FFMS2(self) -> SourceFilter:
        return SourceFilter.FFMS2

    @classproperty
    def LSMASH(self) -> SourceFilter:
        return SourceFilter.LSMASH


SRC_FILE = src_file
FileInfo = src_file


def src(
    filePath: PathLike,
    preview_sourcefilter: SourceFilter | None = SourceFilter.AUTO,
    sourcefilter: SourceFilter = SourceFilter.BESTSOURCE,
    **kwargs: KwargsT,
) -> vs.VideoNode:
    """
    Uses lsmas for previewing and bestsource otherwise.
    Still supports dgi files directly if dgdecodenv is installed to not break existing scripts.

    :param filepath:                Path to video or dgi file
    :param preview_sourcefilter:    Source filter to be used when previewing using vspreview.
                                    `None` will make it fall back to `sourcefilter`.
    :param sourcefilter:            Source filter to be used otherwise.
    :param kwargs:                  Other arguments you may or may not wanna pass to the indexer.
    :return:                        Video Node
    """
    filePath = ensure_path_exists(filePath, src)
    dgiFile = filePath.with_suffix(".dgi")
    if filePath.suffix.lower() == ".dgi" or dgiFile.exists():
        if not hasattr(core, "dgdecodenv"):
            raise error("Trying to use a dgi file without dgdecodenv installed.", src)
        return core.lazy.dgdecodenv.DGSource(str(filePath.resolve()) if not dgiFile.exists() else str(dgiFile.resolve()), **kwargs)

    is_previewing = False
    try:
        from vspreview import is_preview  # type: ignore

        is_previewing = is_preview()
    except:
        debug("Could not check if we're currently previewing. Is vspreview installed?", src)

    force_lsmas = kwargs.pop("force_lsmas", False)
    if force_lsmas:
        warn("force_lsmas is deprecated!\nPlease switch to using the explicit sourcefilter params.", src, 5)
        preview_sourcefilter = None
        sourcefilter = SourceFilter.LSMASH

    force_bs = kwargs.pop("force_bs", False)
    if force_bs:
        warn("force_bs is deprecated!\nPlease switch to using the explicit sourcefilter params.", src, 5)
        preview_sourcefilter = None
        sourcefilter = SourceFilter.BESTSOURCE

    if preview_sourcefilter is SourceFilter.AUTO:
        preview_sourcefilter = SourceFilter.LSMASH if "m2ts" in filePath.name.lower() else SourceFilter.FFMS2

    if sourcefilter is SourceFilter.AUTO:
        sourcefilter = SourceFilter.LSMASH if "m2ts" in filePath.name.lower() else SourceFilter.BESTSOURCE

    if is_previewing and preview_sourcefilter is not None:
        return _call_sourcefilter(filePath.resolve(), preview_sourcefilter, **kwargs)
    else:
        return _call_sourcefilter(filePath.resolve(), sourcefilter, **kwargs)


def _call_sourcefilter(fileIn: Path, sourcefilter: SourceFilter, **kwargs) -> vs.VideoNode:
    filePath = str(fileIn)
    match sourcefilter:
        case SourceFilter.BESTSOURCE:
            if not hasattr(core, "bs"):
                raise error("Bestsource plugin is not installed!", src)
            show_progress = kwargs.pop("showprogress", True)

            info(f"Indexing '{fileIn.name}' using bestsource.", src)
            return core.lazy.bs.VideoSource(filePath, showprogress=show_progress, **kwargs)
        case SourceFilter.LSMASH:
            if not hasattr(core, "lsmas"):
                raise error("LSMASH plugin is not installed!", src)

            info(f"Indexing '{fileIn.name}' using LSMASH LWLibavSource.", src)
            return core.lazy.lsmas.LWLibavSource(filePath, **kwargs)
        case SourceFilter.FFMS2:
            if not hasattr(core, "ffms2"):
                raise error("FFMS2 plugin is not installed!", src)

            info(f"Indexing '{fileIn.name}' using FFMS2.", src)
            return core.lazy.ffms2.Source(filePath, **kwargs)
        case _:
            raise error(f"Invalid sourcefilter passed! ({sourcefilter})", src)


def frames_to_samples(frame: int | None, sample_rate: vs.AudioNode | int = 48000, fps: vs.VideoNode | Fraction = Fraction(24000, 1001)) -> int:
    """
    Converts a frame number to a sample number

    :param frame:           The frame number
    :param sample_rate:     Can be a flat number like 48000 (=48 kHz) or an AudioNode to get the sample rate from
    :param fps:             Can be a Fraction or a VideoNode to get the fps from

    :return:                The sample number
    """
    if not frame:
        return 0
    sample_rate = sample_rate.sample_rate if isinstance(sample_rate, vs.AudioNode) else sample_rate
    fps = Fraction(fps.fps_num, fps.fps_den) if isinstance(fps, vs.VideoNode) else fps
    return int(sample_rate * (fps.denominator / fps.numerator) * frame)


f2s = frames_to_samples


def generate_keyframes(clip: vs.VideoNode, start_frame: int = 0) -> list[int]:
    clip = clip.resize.Bilinear(640, 360, format=vs.YUV410P8)
    clip = clip.wwxd.WWXD()
    if start_frame:
        clip = clip[start_frame:]

    frames = list[int]()
    for i in range(1, clip.num_frames):
        if clip.get_frame(i).props.Scenechange == 1:
            frames.append(i)

    return frames


def generate_qp_file(clip: vs.VideoNode, start_frame: int = 0) -> str:
    filepath = Path(get_workdir(), f"qpfile_{start_frame}.txt")
    temp = Path(get_temp_workdir(), "qpfile.txt")
    if filepath.exists():
        info("Reusing existing QP File.")
        return str(filepath.resolve())
    info("Generating QP File...")

    out = ""
    keyframes = generate_keyframes(clip, start_frame)

    for i in keyframes:
        out += f"{i} I -1\n"

    with open(temp, "w") as file:
        file.write(out)

    temp.rename(filepath)
    clean_temp_files()

    return str(filepath.resolve())
